"use strict"
var assert = require("assert")
var fs = require("fs")
var path = require("path")
var nrrd = require("nrrd-js")
var ndarray = require("ndarray")
var fill = require("ndarray-fill")
var adjacency = require("./lib/adjacency")

var vectorsParsed = nrrd.parse(fs.readFileSync(path.join("out","vectors.nrrd")))
var vectors = ndarray(vectorsParsed.data, vectorsParsed.sizes.slice().reverse())
var shape = vectors.shape

console.log("Gathering data...")
var data = [], opposites = [], index = ndarray(new Int32Array(shape[0]*shape[1]*shape[2]), [shape[0],shape[1],shape[2]])
fill(index, function() { return -1 })
var r, c, pos, vx, vy
for(r=0; r<shape[0]; r++) {
  for(c=0; c<shape[1]; c++) {
    for(pos=0; 2*pos<shape[2]; pos++) {
      vx = vectors.get(r,c,2*pos+0)/255
      vy = vectors.get(r,c,2*pos+1)/255
      if (vx*vx+vy*vy < 1e-3) continue
      data.push([c,r,vx,vy])
      data.push([c,r,-vx,-vy])
      opposites.push(data.length-1)
      opposites.push(data.length-2)
      index.set(r,c,2*pos+0, data.length-2)
      index.set(r,c,2*pos+1, data.length-1)
    }
  }
}

console.log("Gathering adjacency lists...")
var ALs = new Array(data.length)
var i
for(i=0; i<data.length; i++) {
  ALs[i] = adjacency.adjacencyList([data[i][2],data[i][3]])
}

console.log("Finding connections...")
var connections = new Array(ALs.length), AL, nAL, intersection
var r, c, pos, a, npos, nc, nr, ind, nind, vx, vy, nvx, nvy
for(r=0; r<shape[0]; r++) {
  for(c=0; c<shape[1]; c++) {
    for(pos=0; pos<shape[2]; pos++) {
      ind = index.get(r,c,pos)
      if (ind < 0) continue
      vx = data[ind][2]
      vy = data[ind][3]
      connections[ind] = []
      AL = ALs[ind]
      for(a=0; a<AL.length; a++) {
        nc = c+AL[a][0]
        nr = r+AL[a][1]
        if (!(0<=nc && nc<shape[1] &&
              0<=nr && nr<shape[0])) continue
        for(npos=0; npos<shape[2]; npos++) {
          nind = index.get(nr,nc,npos)
          if (nind < 0) continue
          nAL = ALs[nind]
          intersection = adjacency.intersectALs([AL[a]],nAL)
          if (intersection.length === 0) continue
          nvx = data[nind][2]
          nvy = data[nind][3]
          if (Math.pow(vx*nvx+vy*nvy,2) < 0.75*(vx*vx+vy*vy)*(nvx*nvx+nvy*nvy)) continue
          connections[ind].push(nind)
        }
      }
    }
  }
}

// Check connectivity data
console.log("Checking data...")
var i, cs, j, c, cso
var thresholdedCount = 0
for(i=0; i<connections.length; i++) {
  cs = connections[i]
  for(j=0; j<cs.length; j++) {
    c = cs[j]
    cso = connections[opposites[c]]
    assert(cso.indexOf(opposites[i])>=0, "i -> c, but not o(c) -> o(i)!" +
                                         "\ni,c,o(c),o(i): " + [i, c, opposites[c], opposites[i]] +
                                         "\ndata[i]: " + data[i] +
                                         "\ndata[c]: " + data[c] +
                                         "\ndata[o(c)]: " + data[opposites[c]] +
                                         "\ndata[o(i)]: " + data[opposites[i]] +
                                         "\ncs: " + cs +
                                         "\ncso: " + cso +
                                         "\nALs[i]: " + ALs[i] +
                                         "\nALs[c]: " + ALs[c] +
                                         "\nALs[o(c)]: " + ALs[opposites[c]] +
                                         "\nALs[o(i)]: " + ALs[opposites[i]])
  }
}

fs.writeFileSync(path.join("out","graph.json"), JSON.stringify({data: data, connections: connections, opposites: opposites, sizes: Array.apply([], shape).slice(0,-1).reverse()}))

