"use strict"
var assert = require("assert")

module.exports.adjacencyList = adjacencyList
module.exports.perpAdjacencyList = perpAdjacencyList
module.exports.directionAdjacencies = directionAdjacencies
module.exports.directionPerpAdjacencies = directionPerpAdjacencies
module.exports.intersectALs = intersectALs

// scale is assumed to contain positive values corresponding to the point to point distances in the primary directions.
function adjacencyList(orientation_org, scale) {
  var j, d = orientation_org.length, points = []
  if (scale === undefined) {
    scale = new Array(d)
    for(j=0; j<d; j++) {
      scale[j] = 1
    }
  }
  
  // Compute scaled orientation
  var orientation = new Array(d), invlenO = 0
  for(j=0; j<d; j++) {
    orientation[j] = orientation_org[j] / scale[j]
    invlenO += orientation[j]*orientation[j]
  }
  invlenO = 1.0/Math.sqrt(invlenO)
  for(j=0; j<d; j++) {
    orientation[j] = orientation[j] * invlenO
  }
  
  // Iterate over neighbours in lexicographical order, and only add those that are within the "half-cylinder" around the orientation
  var numPoints = Math.pow(3, d)|0, p, pr, point, dotO, perpO
  for(p=0; p<numPoints; p++) {
    if (2*p===numPoints-1) continue // Center (0,0,...)
    // Form point
    point = new Array(d)
    pr = p
    for(j=d; j-->0;) {
      point[j] = (pr % 3) - 1
      pr = (pr / 3) | 0
    }
    // Compute dot product with orientation
    dotO = 0
    for(j=0; j<d; j++) {
      dotO += orientation[j]*point[j]
    }
    if (dotO <= 0) continue // On the "wrong side"
    // Compute component perpendicular to orientation
    perpO = 0
    for(j=0; j<d; j++) {
      perpO += Math.pow(point[j]-orientation[j]*dotO, 2)
    }
    if (perpO > 0.65) continue // Too far from centre axis (note that perpO is the square of the magnitude of the perpendicular component)
    // Put point in list
    points.push(point)
  }
  
  // Remove duplicates
  /*points.sort(lexcomp)
  var q
  for(p=1,q=1; p<points.length; p++) {
    if (lexcomp(points[p], points[q-1]) !== 0) {
      points[q] = points[p]
      q++
    }
  }
  points.length = Math.min(points.length, q)*/
  
  return points
}
function perpAdjacencyList(orientation_org, scale) {
  var j, d = orientation_org.length, points = []
  if (scale === undefined) {
    scale = new Array(d)
    for(j=0; j<d; j++) {
      scale[j] = 1
    }
  }
  
  // Compute scaled orientation
  var orientation = new Array(d), invlenO = 0
  for(j=0; j<d; j++) {
    orientation[j] = orientation_org[j] / scale[j]
    invlenO += orientation[j]*orientation[j]
  }
  invlenO = 1.0/Math.sqrt(invlenO)
  for(j=0; j<d; j++) {
    orientation[j] = orientation[j] * invlenO
  }
  
  // Iterate over neighbours in lexicographical order, and only add those that are within the "half-cylinder" around the orientation
  var numPoints = Math.pow(3, d)|0, p, pr, point, dotO, perpO
  for(p=0; p<numPoints; p++) {
    if (2*p===numPoints-1) continue // Center (0,0,...)
    // Form point
    point = new Array(d)
    pr = p
    for(j=d; j-->0;) {
      point[j] = (pr % 3) - 1
      pr = (pr / 3) | 0
    }
    // Compute dot product with orientation
    dotO = 0
    for(j=0; j<d; j++) {
      dotO += orientation[j]*point[j]
    }
    if (dotO > Math.sqrt(0.65)) continue // Too far from (hyper)plane perpendicular to orientation
    // Put point in list
    points.push(point)
  }
  
  // Remove duplicates
  /*points.sort(lexcomp)
  var q
  for(p=1,q=1; p<points.length; p++) {
    if (lexcomp(points[p], points[q-1]) !== 0) {
      points[q] = points[p]
      q++
    }
  }
  points.length = Math.min(points.length, q)*/
  
  return points
}
/*
function adjacencyList(orientation, scale) {
  var i, j, d = orientation.length, points = []
  if (scale === undefined) {
    scale = new Array(d)
    for(j=0; j<d; j++) {
      scale[j] = 1
    }
  }
  
  // Determine which side of the box around the origin is closest.
  var closestIntersection = Infinity, intersection = new Array(d)
  for(i=0; i<d; i++) {
    intersection[i] = scale[i]/Math.abs(orientation[i])
    if (intersection[i] < closestIntersection) {
      closestIntersection = intersection[i]
    }
  }
  
  // Find cell on side of box we just found (or any that is similarly close)
  var range = new Array(d), r, numPoints, p, pr, point
  for(i=0; i<d; i++) {
    if (intersection[i] < closestIntersection+1e-2) {
      numPoints = 1
      for(j=0; j<d; j++) {
        if (j === i) {
          range[j] = orientation[j]<0 ? [-1,1] : [1,1]
        } else {
          range[j] = orientation[j]<-1e-2 ? [-1,2] : orientation[j]>1e-2 ? [0,2] : [-1,3]
        }
        numPoints *= range[j][1] // range[j][1] contains the /number of/ positions in the range
      }
      for(p=0; p<numPoints; p++) {
        pr = p
        point = new Array(d)
        for(j=0; j<d; j++) {
          r = range[j]
          point[j] = r[0] + (pr % r[1])
          pr = (pr / r[1]) | 0
        }
        points.push(point)
      }
    }
  }
  
  // Remove duplicates
  points.sort(lexcomp)
  var q
  for(p=1,q=1; p<points.length; p++) {
    if (lexcomp(points[p], points[q-1]) !== 0) {
      points[q] = points[p]
      q++
    }
  }
  points.length = Math.min(points.length, q)
  
  return points
}

// scale is assumed to contain positive values corresponding to the point to point distances in the primary directions.
function perpAdjacencyList(orientation, scale) {
  var i, j, d = orientation.length, points = []
  if (scale === undefined) {
    scale = new Array(d)
    for(j=0; j<d; j++) {
      scale[j] = 1
    }
  }
  
  // Determine which side of the box around the origin is closest.
  var closestIntersection = Infinity, intersection = new Array(d)
  for(i=0; i<d; i++) {
    intersection[i] = scale[i]/Math.abs(orientation[i])
    if (intersection[i] < closestIntersection) {
      closestIntersection = intersection[i]
    }
  }
  
  // Find "directly" connected points along a (hyper)plane parallel to the closest box side (or multiple (hyper)planes if multiple box sides are similarly close)
  var range = new Array(d), r, numPoints, p, pr, point
  for(i=0; i<d; i++) {
    if (intersection[i] < closestIntersection+1e-2) {
      for(p=0; p<d; p++) {
        if (p === i) continue
        point = new Array(d)
        for(j=0; j<d; j++) {
          point[j] = j === p ? 1 : 0
        }
        points.push(point)
        point = new Array(d)
        for(j=0; j<d; j++) {
          point[j] = j === p ? -1 : 0
        }
        points.push(point)
      }
    }
  }
  
  // Remove duplicates
  points.sort(lexcomp)
  var q
  for(p=1,q=1; p<points.length; p++) {
    if (lexcomp(points[p], points[q-1]) !== 0) {
      points[q] = points[p]
      q++
    }
  }
  points.length = Math.min(points.length, q)
  
  return points
}*/

// Returns a (lex.) sorted list of neighbouring positions and orientations for each given direction
// TODO: Turn directions into a list of lists?
function directionAdjacencies(directions, scale) {
  var i, j, o, p, a, adjI, adjJ, adjC
  var adjList = new Array(directions.length/3), adjL
  var adjPosLists = new Array(adjList.length)
  for(i=0, o=0; i<directions.length; i+=3, o+=1) { // Cache position adjacency lists for all orientations
    adjPosLists[o] = adjacencyList([directions[0+i], directions[1+i], directions[2+i]], scale)
  }
  for(o=0; o<adjPosLists.length; o+=1) {
    adjI = adjPosLists[o]
    adjL = []
    for(p=0; p<adjPosLists.length; p+=1) {
      adjJ = adjPosLists[p]
      // Compute the intersection of the sets of adjacent positions for both orientations
      // For every direction in the intersection, it is true that if from the current position and orientation (i) we connect to the new position and orientation j, we can also get back when walking along the opposite direction.
      adjC = intersectALs(adjI, adjJ)
      // Add the combinations of positions in adjC and the current orientation to adjacencyList
      for(a=0; a<adjC.length; a++) {
        adjL.push([adjC[a][0],adjC[a][1],adjC[a][2], p])
      }
    }
    adjL.sort(lexcomp)
    adjList[o] = adjL
  }
  // TODO: check symmetry: if (0,o1) points to (offset,o2), then (0,-o2) should point to (-offset,-o1).
  return adjList
}

function directionPerpAdjacencies(directions, scale) {
  var i, j, o, p, a, adjI, padjI, adjJ, padjJ, adjC, padjC
  var adjList = new Array(directions.length/3), adjL
  var adjPosLists = new Array(adjList.length)
  var adjPerpPosLists = new Array(adjList.length)
  for(i=0, o=0; i<directions.length; i+=3, o+=1) { // Cache position adjacency lists for all orientations
    adjPosLists[o] = adjacencyList([directions[0+i], directions[1+i], directions[2+i]], scale)
    adjPerpPosLists[o] = perpAdjacencyList([directions[0+i], directions[1+i], directions[2+i]], scale)
  }
  for(o=0; o<adjPosLists.length; o+=1) {
    adjI = adjPosLists[o]
    padjI = adjPerpPosLists[o]
    adjL = []
    for(p=0; p<adjPosLists.length; p+=1) {
      adjJ = adjPosLists[p]
      // Compute the intersection of the sets of adjacent positions for both orientations
      // For every direction in the intersection, it is true that if from the current position and orientation (i) we connect to the new position and orientation j, we can also get back when walking along the opposite direction.
      adjC = intersectALs(adjI, adjJ)
      if (adjC.length === 0) continue // Skip if they don't have any "parallel" directions in common.
      // Compute the union of the perpendicular directions
      padjJ = adjPerpPosLists[p]
      padjC = unionALs(padjI, padjJ)
      // Add the combinations of positions in padjC and the current orientation to adjacencyList
      for(a=0; a<padjC.length; a++) {
        adjL.push([padjC[a][0],padjC[a][1],padjC[a][2], p])
      }
    }
    adjL.sort(lexcomp)
    adjList[o] = adjL
  }
  // TODO: check symmetry: if (0,o1) points to (offset,o2), then (0,-o2) should point to (-offset,-o1).
  return adjList
}

// Uses the fact that the lists are sorted (and do not contain duplicates)
function intersectALs(al1, al2) {
  var p=0, q=0, alI = [], comp
  while(p<al1.length && q<al2.length) {
    comp = lexcomp(al1[p], al2[q])
    if (comp < 0) {
      p++
    } else if (comp > 0) {
      q++
    } else {
      alI.push(al1[p])
      p++
      q++
    }
  }
  return alI
}

// Uses the fact that the lists are sorted (and do not contain duplicates)
function unionALs(al1, al2) {
  var p=0, q=0, alU = [], comp
  while(p<al1.length && q<al2.length) {
    comp = lexcomp(al1[p], al2[q])
    if (comp < 0) {
      alU.push(al1[p])
      p++
    } else if (comp > 0) {
      alU.push(al2[p])
      q++
    } else {
      alU.push(al1[p])
      p++
      q++
    }
  }
  return alU
}

function lexcomp(a,b) {
  for(var i=0; i<Math.min(a.length,b.length); i++) {
    if (a[i]<b[i]) return -1
    if (a[i]>b[i]) return 1
  }
  if (a.length<b.length) -1
  if (a.length>b.length) 1
  return 0
}
