"use strict"
var assert = require("assert")
var fs = require("fs")
var path = require("path")
var convolve = require("ndarray-convolve")
var pack = require("ndarray-pack")
var ndarray = require("ndarray")
var cwise = require("cwise")

var filters = JSON.parse(fs.readFileSync(path.join(__dirname, "filters.json")))
filters = ["xxxx", "xxxy", "xxyy", "xyyy", "yyyy"].map(function(name){return filters[name]}) // Convert to list rather than object
filters = filters.map(function(kernel){return pack(kernel)}) // Convert arrays of arrays to ndarrays
filters.map(function(kernel){assert(kernel.shape[0]==filters[0].shape[0] && kernel.shape[1]==filters[0].shape[1] && kernel.shape.length==2, "All filters should have the same (2D) size!")})
assert(filters.length == 5, "Just checking that there are indeed as many filters as I expect there to be.")

var spatiallyBound = cwise({ // Make filter spatially bounded (cos^2 window)
  args: ["index", "array", "scalar", "scalar"],
  body: function(i, a, c, r) {
    var dist = Math.sqrt((i[0]-c[0])*(i[0]-c[0]) + (i[1]-c[1])*(i[1]-c[1]))
    dist = dist<r ? (dist/r) : 1
    var coef = Math.cos(0.5*Math.PI*dist)
    coef *= coef
    a *= coef
  }
})
filters.map(function(kernel){spatiallyBound(kernel, [kernel.shape[0]>>1,kernel.shape[1]>>1], 20)})
// TODO: Only keep the non-zero part of the filters (or at least the part within a bounding box around the non-zero part)

function createOrientationScore(img_in) {
  var i, f, score, shape, totalSize = 1, img_out, filter
  shape = Array.apply([], img_in.shape)
  for(i=0; i<shape.length; i++) {
    shape[i] += filters[0].shape[i]
    shape[i] -= 1
    totalSize *= shape[i]
  }
  shape = shape.concat(5)
  totalSize *= 5
  score = ndarray(new Float32Array(totalSize), shape)
  for(f=0; f<5; f++) {
    filter = filters[f]
    img_out = score.pick(null,null,f)
    convolve(img_out, img_in, filter)
  }
  return score
}

// Computes the orientation scores of imgarr, using a tensor of the given order.
// For now we can only handle 2D images and fourth order tensors.
// TODO: Make imgnew to have a data type that is kind of the "join" of the types of the kernels we make and the data type used for imgarr.
module.exports = function(imgarr, order) {
  assert(imgarr.shape.length == 2, "Orientation scores, for now, only support 2D (greyscale) images.")
  assert(order == 4, "For now, only tensorial orientation scores of order four can be built.")

  return createOrientationScore(imgarr)
}

