Path openings
=============
Traditionally, path openings find approximately straight paths in (binary) images. This is done though by constructing directed acyclic graphs on those images and finding paths in those (one graph is constructed for each direction). This library allows doing more or less the same, but on general directed graphs (possibly with cycles).

Since the library actually computes the weights/lengths of the paths rather than the final result of the opening, it can be considered to compute a "path opening transform". The actual path opening would be accomplished by just keeping those nodes where the weight is sufficiently large.

## Example
```javascript
var pathOpening = require("path-opening");

var adjacencyList = [[1, 2], [3], [3], []],
    nodeWeights = [1, 2, 4, 8],
    edgeWeights = [[16, 32], [128], [64], []],
    pot;

pot = pathOpening(adjacencyList, nodeWeights, edgeWeights);
// now pot.maxWeightBefore equals [0, 1+16, 1+32, 1+16+2+128]
//     pot.parent equals [-1, 0, 0, 2]
```

## API
```javascript
var pathOpening = require("path-opening");
```

### pathOpening(adjacencyList, nodeWeights[, edgeWeights])

**Parameters:**

* `adjacencyList`: A list of lists representing the adjacency list of a directed acyclic graph.
* `nodeWeights`: A list of node weights (same length as adjacencyList).
* `edgeWeights`: (optional) A list of edge weights (same shape as adjacencyList).

**Returns:** An object containg:

* `maxWeightBefore`: A list of the weights of the heaviest paths _ending_ in each node (not counting the weight of the final node itself).
* `parent`: A list of pointers to "previous" nodes for reconstructing the heaviest paths (the heaviest path ending in a node can be found by following the parent pointers). A value of -1 means the node is the starting node for such a heaviest path.

Note that this computes the weights of the longest/heaviest paths _ending_ in each node. To compute the weight of the longest path _through_ a node: either make use of some symmetry inherent in the graph, or run pathOpening twice (once on the normal adjacency list and once on the adjacency list corresponding to the graph where each edge is inverted). The weights corresponding to the heaviest paths _through_ each node then correspond to summing the node weights as well as the two weight arrays computed using `pathOpening`.


