"use strict";

var assert = require('assert');

module.exports = function(adjacencyList, nodeWeights, edgeWeights) {
    var v, w, edges, weights, e, myweight, myedgeweight,
        startNodes = [],
        maxWeightBefore = new Array(adjacencyList.length),
        indegree = new Int32Array(adjacencyList.length),
        parent = new Int32Array(adjacencyList.length);

    // Initialize
    for(v=0; v<adjacencyList.length; v++) {
        maxWeightBefore[v] = 0;
        //indegree[v] = 0;
        parent[v] = -1;
    }
    
    // Count in-degree
    for(v=0; v<adjacencyList.length; v++) {
        edges = adjacencyList[v];
        for(e=0; e<edges.length; e++) {
            indegree[edges[e]]++;
        }
    }
    
    // Find nodes with zero in-degree
    for(v=0; v<indegree.length; v++) {
        if (indegree[v] === 0) startNodes.push(v);
    }
    assert(startNodes.length>0, "Given adjacency list does NOT describe an acyclic graph.");
    
    // Find longest paths ending at each node
    // Essentially topological sort, but without outputting the sorting, just updating weights.
    // Note that although a stack is used, nodes are only added once their in-degree is zero, ensuring a node is processed only after all its predecessors have been processed.
    if (edgeWeights === undefined) {
        while(startNodes.length>0) { 
            v = startNodes.pop();
            edges = adjacencyList[v];
            myweight = maxWeightBefore[v] + nodeWeights[v];
            for(e=0; e<edges.length; e++) {
                w = edges[e];
                if (myweight > maxWeightBefore[w]) {
                    maxWeightBefore[w] = myweight;
                    parent[w] = v;
                }
                indegree[w]--;
                if (indegree[w] === 0) startNodes.push(w);
            }
        }
    } else {
        while(startNodes.length>0) { 
            v = startNodes.pop();
            edges = adjacencyList[v];
            weights = edgeWeights[v];
            myweight = maxWeightBefore[v] + nodeWeights[v];
            for(e=0; e<edges.length; e++) {
                myedgeweight = myweight + weights[e];
                w = edges[e];
                if (myedgeweight > maxWeightBefore[w]) {
                    maxWeightBefore[w] = myedgeweight;
                    parent[w] = v;
                }
                indegree[w]--;
                if (indegree[w] === 0) startNodes.push(w);
            }
        }
    }
    
    return {maxWeightBefore: maxWeightBefore, parent: parent};
}

