"use strict"
var assert = require("assert")
var fs = require("fs")
var path = require("path")
var scc = require("strongly-connected-components")
var po = require("path-opening")

var dataConn = JSON.parse(fs.readFileSync(path.join("out","graph.json"), {encoding: "utf-8"}))
var data = dataConn.data
var connections = dataConn.connections
var opposites = dataConn.opposites

// Threshold data
// Note that currently the threshold is zero, so nothing is actually thresholded (we're comparing to vector magnitudes)
console.log("Thresholding...")
var i, cs, j, k, c, threshold = 0.00, maxvalue = 0, norm
var thresholdedCount = 0
for(i=0; i<connections.length; i++) {
  norm = Math.sqrt(data[i][2]*data[i][2]+data[i][3]*data[i][3])
  if (norm < threshold) {
    data[i][2] = 0
    data[i][3] = 0
    connections[i] = []
    thresholdedCount++
  } else {
    cs = connections[i]
    for(j=0,k=0; j<cs.length; j++) {
      c = cs[j]
      if (Math.sqrt(data[c][2]*data[c][2]+data[c][3]*data[c][3]) < threshold) {
        // skip?
      } else {
        cs[k] = c
        k++
      }
    }
    cs.length = k
  }
  if (norm > maxvalue) {
    maxvalue = norm
  }
}
console.log("Kept " + (connections.length - thresholdedCount) + " out of " + connections.length)
console.log("Max. value: " + maxvalue)

fs.writeFileSync(path.join("out","graph-thresholded.json"), JSON.stringify(dataConn))

// Compute strongly connected components
console.log("Strongly connected components...")
var sccRet = scc(dataConn.connections)
var components = sccRet.components
var adjacencyList = sccRet.adjacencyList

var histogram = [], i, s, maxsize = 0, maxind, totalSize = 0
for(i=0; i<components.length; i++) {
  s = components[i].length
  assert(s>=1, "Every strongly connected component should have at least one element in it.")
  totalSize += s
  if (s > maxsize) {
    maxsize = s
    maxind = i
  }
  s = Math.floor(s/1000)
  if (s in histogram) {
    histogram[s] += 1
  } else {
    histogram[s] = 1
  }
}
console.log("" + [totalSize, connections.length])
console.log(histogram)
console.log("Maximum component size: " + maxsize)

/*var maxComponent = components[maxind]
for(i=0; i<connections.length; i++) {
  data[6+7*i] = 0.01
}
for(i=0; i<maxComponent.length; i++) {
  data[6+7*maxComponent[i]] = 1
}
fs.writeFileSync("out_annotated_graph_largest.json", JSON.stringify(dataConn))*/

// Compute scc opposites
console.log("SCC opposites...")
var sccIndex = new Int32Array(connections.length)
var sccOpposites = new Int32Array(components.length)
for(i=0; i<connections.length; i++) {
  sccIndex[i] = -1
}
for(i=0; i<components.length; i++) { // Gather scc indices for all points
  sccOpposites[i] = -1
  cs = components[i]
  for(j=0; j<cs.length; j++) {
    sccIndex[cs[j]] = i
  }
}
for(i=0; i<connections.length; i++) {
  assert(sccIndex[i] !== -1, "Some vertex not included in any scc?!?!")
  if (sccOpposites[sccIndex[i]] === -1) {
    sccOpposites[sccIndex[i]] = sccIndex[opposites[i]]
  } else {
    assert(sccOpposites[sccIndex[i]] === sccIndex[opposites[i]], "Conflicting scc opposites information!? " + [i, opposites[i], sccIndex[i], sccIndex[opposites[i]], sccOpposites[sccIndex[i]]])
  }
}
for(i=0; i<components.length; i++) { // Gather scc indices for all points
  assert(sccOpposites[i] !== -1, "No opposite for scc " + i + "?")
  assert(sccOpposites[sccOpposites[i]] === i, "Inconsistent opposite for scc " + i + "!")
}

// Compute node and edge weights
// TODO: More intelligent node and edge weights?
console.log("Edge weights...")
var i, cs, j, sccNodeWeights = new Array(adjacencyList.length)//, sccEdgeWeights = new Array(adjacencyList.length)
for(i=0; i<adjacencyList.length; i++) {
  cs = adjacencyList[i]
  sccNodeWeights[i] = components[i].length
  /*sccEdgeWeights[i] = []
  for(j=0; j<cs.length; j++) {
    sccEdgeWeights[i].push(0)
  }*/
}

// Compute path opening transform
console.log("Path opening transform...")
var pot = po(adjacencyList, sccNodeWeights/*, sccEdgeWeights*/)
var pathLength = new Array(connections.length)
var maxWeight = 0, myweight
for(i=0; i<pathLength.length; i++) {
  pathLength[i] = 0
}
for(i=0; i<components.length; i++) {
  cs = components[i]
  myweight = pot.maxWeightBefore[i] + sccNodeWeights[i] + pot.maxWeightBefore[sccOpposites[i]]
  if (myweight > maxWeight) {
    maxWeight = myweight
  }
  for(j=0; j<cs.length; j++) {
    pathLength[cs[j]] = myweight
    //pathLength[cs[j]] = sccNodeWeights[i]
  }
}
dataConn.weights = pathLength
console.log("maxWeight: " + maxWeight)

fs.writeFileSync(path.join("out","graph-pot.json"), JSON.stringify(dataConn))

// Path opening "granulometry"
console.log("Path opening granulometry...")
var granulometry = new Int32Array(Math.round(maxWeight)+1)
for(i=0; i<components.length; i++) {
  cs = components[i]
  myweight = Math.round(pathLength[cs[0]])
  granulometry[myweight]++
}
granulometry[1] -= thresholdedCount // Subtract all of the data points that we thresholded.
var granCSVlines = ['"path length","voxel count"']
for(i=0; i<granulometry.length; i++) {
  granCSVlines.push('' + i + ',' + granulometry[i])
}
fs.writeFileSync(path.join("out","granulometry.csv"), granCSVlines.join('\n'))

